/****************************************************************************
 * defragmusc - a program to make graphic files made with the MUSC converter
 * more compact
 *
 * Jan-2020 V0.2
 * Wilfried Elmenreich
 * Code under CC-BY-4.0 license
 *
 * This tool processes files that were generated with Algotech's MUSC converter
 * (released September 2009 on CSDB: https://csdb.dk/release/?id=83041)
 *
 * An original MUSC file has a size of 121 blocks and spreads its code an data
 * in the memory area between $0801 and $5f40
 * defragmusc packs the code parts closer together so that the file has now 79 blocks
 * and lies in memory between $31b7 and $5f40
 *
 * In addition, some small assembler routines are added supporting the use of a MUSC
 * file as a splash screen:
 *
 * SYS 24384 ($5F40) Display pic and go into forever loop (original behavior)
 * SYS 12727 ($31b7) Display pic and return (ROM is banked out -> does not work with BASIC)
 * SYS 24390 ($5F46) Display pic, wait for key and restore VIC and IRQ settings
 * SYS 24412 ($5F5C) Stops displaying and restores VIC and IRQ settings
****************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <getopt.h>
#include <string.h>

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;

// machine code for displaying, endless loop, waiting for keypress and restoring VIC settings
// to be injected starting at address $5f40
const uint8_t suppcode[] = {0x20,0xb7,0x31,0x4c,0x43,0x5f,0x20,0xb7,0x31,0xa9,0x0,0x8d,0x0,
                            0xdc,0xad,0x1,0xdc,0xc9,0xff,0xd0,0xf4,0xad,0x1,0xdc,0xc9,0xff,
							0xf0,0xf9,0x78,0xa9,0x37,0x85,0x1,0xad,0x0,0xdd,0x9,0x3,0x8d,0x0,
							0xdd,0xa9,0x4,0x8d,0x88,0x2,0x20,0x15,0xfd,0x20,0xa0,0xe5,0xa9,0x81,
							0x8d,0xd,0xdc,0xa9,0x1,0x8d,0xe,0xdc,0x58,0x60};

typedef struct { 
 uint8_t data[0x10000];
 int length;
} cbmfile_t;

int VERBOSE=1;
char *progname;
int lowestaddress=1e9,highestaddress=-1;
int opt_ignorestart=0;

void load_file(char* filename,cbmfile_t *filedata) {
  FILE *fp;
  int filestart,address;
  fp = fopen(filename, "rb");
  filestart = fgetc(fp) + 256 * fgetc(fp);
  address=filestart;
  
  if (filestart<lowestaddress) lowestaddress=filestart;
  
  while(1) {
    int c=fgetc(fp);
    if (feof(fp)) break;  
    filedata->data[address++]=c;
  }
  fclose(fp);
  address--;
  if (address>highestaddress) highestaddress=address;
  
  if (VERBOSE) {
      printf("Read file '%s' $%04x-$%04x\n",filename,filestart,address);
   }
return;
}
 
void save_file(cbmfile_t filedata, char* filename, int start, int end) {
   FILE *fp;
   fp = fopen(filename, "wb");
   fputc(start & 0xff,fp);
   fputc(start>>8,fp);
   for (int i=start; i<=end; i++)
      fputc(filedata.data[i],fp);
   fclose(fp);
   if (VERBOSE) {
      printf("Written file '%s' $%04x-$%04x\n",filename,start,end);
   }
}

cbmfile_t defragMUCS(cbmfile_t filedata) {
   uint16_t t; //target address for absolute adresses/jumps

   for (int i=8192;i<=8384;i++) {
      filedata.data[i+7999] = filedata.data[i];
   }
   for (int i=2560;i<=5744;i++) {
      if ((filedata.data[i-1]!=32) || (filedata.data[i+1]!=32) || (filedata.data[i]==93))
         filedata.data[i+10446] = filedata.data[i];
      else {
         t=8192+7999+filedata.data[i];
         filedata.data[i+10446] = t & 0xff;
         filedata.data[i+10447] = t >> 8;
         i++;
      }
   }
   for (int i=2070;i<=2348;i++) {
      if (i==2256) {
         t=12912;
         filedata.data[i+10657] = t & 0xff;
         filedata.data[i+10658] = t >> 8;
         i++;         
      } else if (i==2319) {
         t=13006;
         filedata.data[i+10657] = t & 0xff;
         filedata.data[i+10658] = t >> 8;
         i++;         
      } else     
         filedata.data[i+10657] = filedata.data[i];
   }
   t=2258+10657; //12915 address of ISR
   filedata.data[12759] = t & 0xff;
   filedata.data[12764] = t >> 8;   
   t=2301+10657; //12958 address of ISR2
   filedata.data[12925] = t & 0xff;
   filedata.data[12930] = t >> 8;   
   t=2258+10657; //12915 address of ISR (again)
   filedata.data[12979] = t & 0xff;
   filedata.data[12984] = t >> 8;   
   t=12997; //self-mod-storage for A
   filedata.data[t-1] = 169; //opcode for LDA #
   filedata.data[12916] = t & 0xff;
   filedata.data[12917] = t >> 8;     
   t=12999; //self-mod-storage for X
   filedata.data[t-1] = 162; //opcode for LDX #
   filedata.data[12919] = t & 0xff;
   filedata.data[12920] = t >> 8;     
   t=13001; //self-mod-storage for Y
   filedata.data[t-1] = 160; //opcode for LDY #
   filedata.data[12922] = t & 0xff;
   filedata.data[12923] = t >> 8;     
   filedata.data[13002] = 64; //RTI because we shortened the routine
	
   //replace stackpointer overwriting with NOP
   filedata.data[0x3265]=0xea;
   //replace endless loop with RTS
   filedata.data[0x3270]=0x60;	

   return filedata;
}

uint16_t calc_checksum(cbmfile_t filedata, int start, int end) {
   //using fletcher's checksum
   uint8_t sum1=0,sum2=0;
   for (int i=start; i<=end; i++) {
   	  sum1 += filedata.data[i];
   	  sum2 += sum1;
   }
   return sum1+(sum2 << 8);
}

cbmfile_t injectcode(cbmfile_t filedata, const uint8_t codearray[], int start, int end) {
   for (int i=start; i<=end; i++) {
   	  filedata.data[i]=codearray[i-start];
   }
   return filedata;
}


void usage(char *s) {
   if (s!=NULL) fprintf(stderr,"%s\n\n",s);
	
   fprintf(stderr,"Usage: %s [-o outfile] inputfile\n",progname);
   fprintf(stderr,"inputfile: MUSC prg file to be read\n");     
   fprintf(stderr,"outfile: name of file to write\n");   
   exit(1);
}

int main(int argc, char ** argv) {
  int c;
	
  char *infilename[256];
  char *outfilename="a.out.prg";
  int start=-1;
  int end=-1;
  int fill=0x00;
  cbmfile_t filedata,fixedfile,augmentedfile;
  
  progname=argv[0];
  while ((c=getopt(argc,argv,"o:"))!=EOF)
  {
     switch(c)
     {
     case 'o':
        outfilename=optarg;
        break;         
     default:
        usage(NULL);
     }
  }

  if (argc-optind<1) {
  	usage("Please specifiy a filename to process!");
  	exit(0);
  }
  
  if (argc-optind>1) {
  	usage("Too many arguments!");
  	exit(0);
  }
  
  load_file(argv[optind],&filedata);
  
  uint16_t checksum=calc_checksum(filedata,0x10,0x37ff);
  
  if (checksum != 0x1898) {
  	
  	fprintf(stderr,"This program only works with PRG files generated by the MUSC-Hires Converter (by Algotech, September 2009).\n");
  	if (checksum == 0x3054)
  	  fprintf(stderr,"It looks like this file was generated with MUCSU-Hires Converter, which has a different output.\n");
  	else
  	  fprintf(stderr,"It looks like this file was not generated with the right program.\n");  
  	exit(1);  
  }

  fixedfile=defragMUCS(filedata);
  
  //add code for waitkey and restore VIC settings
  augmentedfile=injectcode(fixedfile, suppcode, 0x5f40, 0x5f40+sizeof(suppcode));
  
  printf("Saving file %s from $%04x to $%04x\n",outfilename,0x31b7,0x7f40);

  save_file(augmentedfile,outfilename,0x31b7,0x7f40);
    
  printf("Defragmentation done. Note the following entry points to display the graphics:\n");
  printf("  SYS 24384 ($5F40) Display pic and go into forever loop (original behavior)\n");
  printf("  SYS 12727 ($31b7) Display pic and return (Caution! ROM is banked out -> does not work with BASIC)\n");
  printf("  SYS 24390 ($5F46) Display pic, wait for key and restore VIC and IRQ settings\n");
  printf("  SYS 24412 ($5F5C) Stops displaying and restores VIC and IRQ settings\n");
  
  return 0;
}


